from flask import current_app, request, g
import glob
import random
import re
import json
from ipaddress import IPv4Address, IPv6Address
from hiddifypanel.cache import cache
from hiddifypanel.models import Proxy, ProxyProto, ProxyL3, ProxyTransport, ProxyCDN, Domain, DomainType, ConfigEnum, hconfig, get_hconfigs
from hiddifypanel import hutils


def get_ssh_hostkeys(dojson=False) -> list[str] | str:
    key_files = glob.glob(current_app.config['HIDDIFY_CONFIG_PATH'] + "/other/ssh/host_key/*_key.pub")
    host_keys = []
    for file_name in key_files:
        with open(file_name, "r") as f:
            host_key = f.read().strip()
            host_key = host_key.split()
            if len(host_key) > 2:
                host_key = host_key[:2]  # strip the hostname part
            host_key = " ".join(host_key)
            host_keys.append(host_key)
    if dojson:
        return json.dumps(host_keys)
    return host_keys


def is_proxy_valid(proxy: Proxy, domain_db: Domain, port: int) -> dict | None:
    name = proxy.name
    l3 = proxy.l3
    if not port:
        return {'name': name, 'msg': "port not defined", 'type': 'error', 'proto': proxy.proto}
    if "reality" not in l3 and domain_db.mode == DomainType.reality:
        return {'name': name, 'msg': "reality proxy not in reality domain", 'type': 'debug', 'proto': proxy.proto}

    if "reality" in l3 and domain_db.mode != DomainType.reality:
        return {'name': name, 'msg': "reality proxy not in reality domain", 'type': 'debug', 'proto': proxy.proto}

    if "reality" in l3 and domain_db.grpc and ProxyTransport.grpc != proxy.transport:
        return {'name': name, 'msg': "reality proxy not in reality domain", 'type': 'debug', 'proto': proxy.proto}

    if "reality" in l3 and (not domain_db.grpc) and ProxyTransport.grpc == proxy.transport:
        return {'name': name, 'msg': "reality proxy not in reality domain", 'type': 'debug', 'proto': proxy.proto}

    is_cdn = ProxyCDN.CDN == proxy.cdn or ProxyCDN.Fake == proxy.cdn
    if is_cdn and domain_db.mode not in [DomainType.cdn, DomainType.auto_cdn_ip, DomainType.worker]:
        # print("cdn proxy not in cdn domain", domain, name)
        return {'name': name, 'msg': "cdn proxy not in cdn domain", 'type': 'debug', 'proto': proxy.proto}

    if not is_cdn and domain_db.mode in [DomainType.cdn, DomainType.auto_cdn_ip, DomainType.worker]:
        # print("not cdn proxy  in cdn domain", domain, name, proxy.cdn)
        return {'name': name, 'msg': "not cdn proxy  in cdn domain", 'type': 'debug', 'proto': proxy.proto}

    if proxy.cdn == ProxyCDN.relay and domain_db.mode not in [DomainType.relay]:
        return {'name': name, 'msg': "relay proxy not in relay domain", 'type': 'debug', 'proto': proxy.proto}

    if proxy.cdn != ProxyCDN.relay and domain_db.mode in [DomainType.relay]:
        return {'name': name, 'msg': "relay proxy not in relay domain", 'type': 'debug', 'proto': proxy.proto}

    if domain_db.mode == DomainType.worker and proxy.transport == ProxyTransport.grpc:
        return {'name': name, 'msg': "worker does not support grpc", 'type': 'debug', 'proto': proxy.proto}

    if domain_db.mode != DomainType.old_xtls_direct and "tls" in proxy.l3 and proxy.cdn == ProxyCDN.direct and proxy.transport in [ProxyTransport.tcp, ProxyTransport.XTLS]:
        return {'name': name, 'msg': "only  old_xtls_direct  support this", 'type': 'debug', 'proto': proxy.proto}

    if proxy.proto == "trojan" and not is_tls(l3):
        return {'name': name, 'msg': "trojan but not tls", 'type': 'warning', 'proto': proxy.proto}

    if l3 == "http" and ProxyTransport.XTLS in proxy.transport:
        return {'name': name, 'msg': "http and xtls???", 'type': 'warning', 'proto': proxy.proto}

    if l3 == "http" and proxy.proto in [ProxyProto.ss, ProxyProto.ssr]:
        return {'name': name, 'msg': "http and ss or ssr???", 'type': 'warning', 'proto': proxy.proto}


def get_port(proxy: Proxy, hconfigs: dict, domain_db: Domain, ptls: int, phttp: int, pport: int | None) -> int:
    l3 = proxy.l3
    port = 443
    if isinstance(phttp, str):
        phttp = int(phttp) if phttp != "None" else None  # type: ignore
    if isinstance(ptls, str):
        ptls = int(ptls) if ptls != "None" else None  # type: ignore
    if l3 == "kcp":
        port = hconfigs[ConfigEnum.kcp_ports].split(",")[0]
    elif proxy.proto == ProxyProto.wireguard:
        port = hconfigs[ConfigEnum.wireguard_port]
    elif proxy.proto == "tuic":
        port = domain_db.internal_port_tuic
    elif proxy.proto == "hysteria2":
        port = domain_db.internal_port_hysteria2
    elif l3 == 'ssh':
        port = hconfigs[ConfigEnum.ssh_server_port]
    elif is_tls(l3):
        port = ptls
    elif l3 == "http":
        port = phttp
    else:
        port = int(pport)  # type: ignore
    return port


def is_tls(l3) -> bool:
    return 'tls' in l3 or "reality" in l3


@cache.cache(ttl=300)
def get_proxies(child_id: int = 0, only_enabled=False) -> list['Proxy']:
    proxies = Proxy.query.filter(Proxy.child_id == child_id).all()
    proxies = [c for c in proxies if 'restls' not in c.transport]

    if not hconfig(ConfigEnum.tuic_enable, child_id):
        proxies = [c for c in proxies if c.proto != ProxyProto.tuic]
    if not hconfig(ConfigEnum.wireguard_enable, child_id):
        proxies = [c for c in proxies if c.proto != ProxyProto.wireguard]
    if not hconfig(ConfigEnum.ssh_server_enable, child_id):
        proxies = [c for c in proxies if c.proto != ProxyProto.ssh]
    if not hconfig(ConfigEnum.hysteria_enable, child_id):
        proxies = [c for c in proxies if c.proto != ProxyProto.hysteria2]
    if not hconfig(ConfigEnum.shadowsocks2022_enable, child_id):
        proxies = [c for c in proxies if 'shadowsocks' != c.transport]

    if not hconfig(ConfigEnum.ssfaketls_enable, child_id):
        proxies = [c for c in proxies if 'faketls' != c.transport]
    if not hconfig(ConfigEnum.v2ray_enable, child_id):
        proxies = [c for c in proxies if 'v2ray' != c.proto]
    if not hconfig(ConfigEnum.shadowtls_enable, child_id):
        proxies = [c for c in proxies if c.transport != 'shadowtls']
    if not hconfig(ConfigEnum.ssr_enable, child_id):
        proxies = [c for c in proxies if 'ssr' != c.proto]
    if not hconfig(ConfigEnum.vmess_enable, child_id):
        proxies = [c for c in proxies if 'vmess' not in c.proto]
    if not hconfig(ConfigEnum.vless_enable, child_id):
        proxies = [c for c in proxies if 'vless' not in c.proto]
    if not hconfig(ConfigEnum.trojan_enable, child_id):
        proxies = [c for c in proxies if 'trojan' not in c.proto]
    if not hconfig(ConfigEnum.httpupgrade_enable, child_id):
        proxies = [c for c in proxies if ProxyTransport.httpupgrade not in c.transport]
    if not hconfig(ConfigEnum.ws_enable, child_id):
        proxies = [c for c in proxies if ProxyTransport.WS not in c.transport]
    if not hconfig(ConfigEnum.xtls_enable, child_id):
        proxies = [c for c in proxies if ProxyTransport.XTLS not in c.transport]
    if not hconfig(ConfigEnum.grpc_enable, child_id):
        proxies = [c for c in proxies if ProxyTransport.grpc not in c.transport]
    if not hconfig(ConfigEnum.tcp_enable, child_id):
        proxies = [c for c in proxies if 'tcp' not in c.transport]
    if not hconfig(ConfigEnum.h2_enable, child_id):
        proxies = [c for c in proxies if 'h2' not in c.transport and c.l3 not in [ProxyL3.tls_h2_h1, ProxyL3.tls_h2]]
    if not hconfig(ConfigEnum.kcp_enable, child_id):
        proxies = [c for c in proxies if 'kcp' not in c.l3]
    if not hconfig(ConfigEnum.reality_enable, child_id):
        proxies = [c for c in proxies if 'reality' not in c.l3]
    if not hconfig(ConfigEnum.quic_enable, child_id):
        proxies = [c for c in proxies if 'h3_quic' not in c.l3]
    if not hconfig(ConfigEnum.http_proxy_enable, child_id):
        proxies = [c for c in proxies if 'http' != c.l3]

    if not Domain.query.filter(Domain.mode.in_([DomainType.cdn, DomainType.auto_cdn_ip])).first():
        proxies = [c for c in proxies if c.cdn != "CDN"]

    if not Domain.query.filter(Domain.mode.in_([DomainType.relay])).first():
        proxies = [c for c in proxies if c.cdn != ProxyCDN.relay]

    if not Domain.query.filter(Domain.mode.in_([DomainType.cdn, DomainType.auto_cdn_ip]), Domain.servernames != "", Domain.servernames != Domain.domain).first():
        proxies = [c for c in proxies if 'Fake' not in c.cdn]
    proxies = [c for c in proxies if not ('vless' == c.proto and ProxyTransport.tcp == c.transport and c.cdn == ProxyCDN.direct)]

    if only_enabled:
        proxies = [p for p in proxies if p.enable]
    return proxies


def get_valid_proxies(domains: list[Domain]) -> list[dict]:
    allp = []
    allphttp = [p for p in request.args.get("phttp", "").split(',') if p]
    allptls = [p for p in request.args.get("ptls", "").split(',') if p]
    added_ip = {}
    configsmap = {}
    proxeismap = {}
    for domain in domains:
        if domain.child_id not in configsmap:
            configsmap[domain.child_id] = get_hconfigs(domain.child_id)
            proxeismap[domain.child_id] = get_proxies(domain.child_id, only_enabled=True)
        hconfigs = configsmap[domain.child_id]

        ip = hutils.network.get_domain_ip(domain.domain, version=4)
        ip6 = hutils.network.get_domain_ip(domain.domain, version=6)
        ips = [x for x in [ip, ip6] if x is not None]
        for proxy in proxeismap[domain.child_id]:
            noDomainProxies = False
            if proxy.proto in [ProxyProto.ssh, ProxyProto.wireguard]:
                noDomainProxies = True
            if proxy.proto in [ProxyProto.ss] and proxy.transport not in [ProxyTransport.grpc, ProxyTransport.h2, ProxyTransport.WS, ProxyTransport.httpupgrade]:
                noDomainProxies = True
            options = []
            key = f'{proxy.proto}{proxy.transport}{proxy.cdn}{proxy.l3}'
            if key not in added_ip:
                added_ip[key] = {}
            if proxy.proto in [ProxyProto.ssh, ProxyProto.tuic, ProxyProto.hysteria2, ProxyProto.wireguard, ProxyProto.ss]:
                if noDomainProxies and all([x in added_ip[key] for x in ips]):
                    continue

                for x in ips:
                    added_ip[key][x] = 1

                if proxy.proto in [ProxyProto.ssh, ProxyProto.wireguard, ProxyProto.ss]:
                    if domain.mode == 'fake':
                        continue
                    if proxy.proto in [ProxyProto.ssh]:
                        options = [{'pport': hconfigs[ConfigEnum.ssh_server_port]}]
                    elif proxy.proto in [ProxyProto.wireguard]:
                        options = [{'pport': hconfigs[ConfigEnum.wireguard_port]}]
                    elif proxy.transport in [ProxyTransport.shadowsocks]:
                        options = [{'pport': hconfigs[ConfigEnum.shadowsocks2022_port]}]
                    elif proxy.proto in [ProxyProto.ss]:
                        options = [{'pport': 443}]
                elif proxy.proto == ProxyProto.tuic:
                    options = [{'pport': hconfigs[ConfigEnum.tuic_port]}]
                elif proxy.proto == ProxyProto.hysteria2:
                    options = [{'pport': hconfigs[ConfigEnum.hysteria_port]}]
            else:
                protos = ['http', 'tls'] if hconfigs.get(ConfigEnum.http_proxy_enable) else ['tls']
                for t in protos:
                    for port in hconfigs[ConfigEnum.http_ports if t == 'http' else ConfigEnum.tls_ports].split(','):
                        phttp = port if t == 'http' else None
                        ptls = port if t == 'tls' else None
                        if phttp and len(allphttp) and phttp not in allphttp:
                            continue
                        if ptls and len(allptls) and ptls not in allptls:
                            continue
                        options.append({'phttp': phttp, 'ptls': ptls})

            for opt in options:
                pinfo = make_proxy(hconfigs, proxy, domain, **opt)
                if 'msg' not in pinfo:
                    allp.append(pinfo)
    return allp


def make_proxy(hconfigs: dict, proxy: Proxy, domain_db: Domain, phttp=80, ptls=443, pport: int | None = None) -> dict:

    l3 = proxy.l3
    domain = domain_db.domain
    child_id = domain_db.child_id
    name = proxy.name
    port = hutils.proxy.get_port(proxy, hconfigs, domain_db, ptls, phttp, pport)

    if val_res := hutils.proxy.is_proxy_valid(proxy, domain_db, port):
        # print(val_res)
        return val_res

    if 'reality' in proxy.l3:
        alpn = "h2" if proxy.transport in ['h2', "grpc"] else 'http/1.1'
    else:
        alpn = "h2" if proxy.l3 in ['tls_h2'] or proxy.transport in ["grpc", 'h2'] else 'h2,http/1.1' if proxy.l3 == 'tls_h2_h1' else "http/1.1"
    cdn_forced_host = domain_db.cdn_ip or (domain_db.domain if domain_db.mode != DomainType.reality else hutils.network.get_direct_host_or_ip(4))
    is_cdn = ProxyCDN.CDN == proxy.cdn or ProxyCDN.Fake == proxy.cdn
    base = {
        'name': name,
        'cdn': is_cdn,
        'mode': "CDN" if is_cdn else "direct",
        'l3': l3,
        'host': domain,
        'port': port,
        'server': cdn_forced_host,
        'sni': domain_db.servernames if is_cdn and domain_db.servernames else domain,
        'uuid': str(g.account.uuid),
        'proto': proxy.proto,
        'transport': proxy.transport,
        'proxy_path': hconfigs[ConfigEnum.proxy_path],
        'alpn': alpn,
        'extra_info': f'{domain_db.alias or domain}',
        'fingerprint': hconfigs[ConfigEnum.utls],
        'allow_insecure': domain_db.mode == DomainType.fake or "Fake" in proxy.cdn,
        'dbe': proxy,
        'dbdomain': domain_db
    }
    if proxy.proto in ['tuic', 'hysteria2']:
        base['alpn'] = "h3"
        if proxy.proto == 'hysteria2':
            base['hysteria_up_mbps'] = hconfigs.get(ConfigEnum.hysteria_up_mbps)
            base['hysteria_down_mbps'] = hconfigs.get(ConfigEnum.hysteria_down_mbps)
            base['hysteria_obfs_enable'] = hconfigs.get(ConfigEnum.hysteria_obfs_enable)
            base['hysteria_obfs_password'] = hconfigs.get(ConfigEnum.proxy_path)  # TODO: it should not be correct
        return base
    if proxy.proto in ['wireguard']:
        base['wg_pub'] = g.account.wg_pub
        base['wg_pk'] = g.account.wg_pk
        base['wg_psk'] = g.account.wg_psk
        base['wg_ipv4'] = hutils.network.add_number_to_ipv4(hconfigs[ConfigEnum.wireguard_ipv4], g.account.id)
        base['wg_ipv6'] = hutils.network.add_number_to_ipv6(hconfigs[ConfigEnum.wireguard_ipv6], g.account.id)
        base['wg_server_pub'] = hconfigs[ConfigEnum.wireguard_public_key]
        base['wg_noise_trick'] = hconfigs[ConfigEnum.wireguard_noise_trick]
        return base

    if proxy.proto in [ProxyProto.vmess]:
        base['cipher'] = "chacha20-poly1305"

    if l3 in ['reality']:
        base['reality_short_id'] = random.sample(hconfigs[ConfigEnum.reality_short_ids].split(','), 1)[0]
        # base['flow']="xtls-rprx-vision"
        base['reality_pbk'] = hconfigs[ConfigEnum.reality_public_key]
        if (domain_db.servernames):
            all_servernames = re.split('[ \t\r\n;,]+', domain_db.servernames)
            base['sni'] = random.sample(all_servernames, 1)[0]
            if hconfigs[ConfigEnum.core_type] == "singbox":
                base['sni'] = all_servernames[0]
        else:
            base['sni'] = domain_db.domain

        del base['host']
        if base.get('fingerprint'):
            base['fingerprint'] = hconfigs[ConfigEnum.utls]
        # if not domain_db.cdn_ip:
        #     base['server']=hiddify.get_domain_ip(base['server'])

    if "Fake" in proxy.cdn:
        if not hconfigs[ConfigEnum.domain_fronting_domain]:
            return {'name': name, 'msg': "no domain_fronting_domain", 'type': 'debug', 'proto': proxy.proto}
        if l3 == "http" and not hconfigs[ConfigEnum.domain_fronting_http_enable]:
            return {'name': name, 'msg': "no http in domain_fronting_domain", 'type': 'debug', 'proto': proxy.proto}
        if l3 == "tls" and not hconfigs[ConfigEnum.domain_fronting_tls_enable]:
            return {'name': name, 'msg': "no tls in domain_fronting_domain", 'type': 'debug', 'proto': proxy.proto}
        base['server'] = hconfigs[ConfigEnum.domain_fronting_domain]
        base['sni'] = hconfigs[ConfigEnum.domain_fronting_domain]
        # base["host"]=domain
        base['mode'] = 'Fake'
    elif l3 == "http" and not hconfigs[ConfigEnum.http_proxy_enable]:
        return {'name': name, 'msg': "http but http is disabled ", 'type': 'debug', 'proto': proxy.proto}

    path = {
        'vless': f'{hconfigs[ConfigEnum.path_vless]}',
        'trojan': f'{hconfigs[ConfigEnum.path_trojan]}',
        'vmess': f'{hconfigs[ConfigEnum.path_vmess]}',
        'ss': f'{hconfigs[ConfigEnum.path_ss]}',
        'v2ray': f'{hconfigs[ConfigEnum.path_ss]}'
    }

    if base["proto"] in ['v2ray', 'ss', 'ssr']:
        base['cipher'] = hconfigs[ConfigEnum.shadowsocks2022_method]
        base['password'] = f'{hutils.encode.do_base_64(hconfigs[ConfigEnum.shared_secret].replace("-",""))}:{hutils.encode.do_base_64(g.account.uuid.replace("-",""))}'

    if base['proto'] == 'trojan':
        base['password'] = base['uuid']
    if base["proto"] == "ssr":
        base["ssr-obfs"] = "tls1.2_ticket_auth"
        base["ssr-protocol"] = "auth_sha1_v4"
        base["fakedomain"] = hconfigs[ConfigEnum.ssr_fakedomain]
        base["mode"] = "FakeTLS"
        return base
    elif "faketls" in proxy.transport:
        base['fakedomain'] = hconfigs[ConfigEnum.ssfaketls_fakedomain]
        base['mode'] = 'FakeTLS'
        return base
    elif "shadowtls" in proxy.transport:
        base['fakedomain'] = hconfigs[ConfigEnum.shadowtls_fakedomain]
        # base['sni'] = hconfigs[ConfigEnum.shadowtls_fakedomain]
        base['shared_secret'] = hconfigs[ConfigEnum.shared_secret]
        base['mode'] = 'ShadowTLS'
        return base
    elif "shadowsocks" in proxy.transport:
        return base
    if ProxyTransport.XTLS in proxy.transport:
        base['flow'] = 'xtls-rprx-vision'
        return {**base, 'transport': 'tcp'}

    if proxy.proto in {'vless', 'trojan', 'vmess'} and hconfigs.get(ConfigEnum.mux_enable):
        if hconfigs[ConfigEnum.mux_enable]:
            base['mux_enable'] = hconfigs[ConfigEnum.core_type]

            base['mux_protocol'] = hconfigs.get(ConfigEnum.mux_protocol, "h2mux")
            base['mux_max_connections'] = hconfigs.get(ConfigEnum.mux_max_connections, 0)
            base['mux_min_streams'] = hconfigs.get(ConfigEnum.mux_min_streams, 0)
            base['mux_max_streams'] = hconfigs.get(ConfigEnum.mux_max_streams, 0)
            base['mux_padding_enable'] = hconfigs.get(ConfigEnum.mux_padding_enable)

            if hconfigs[ConfigEnum.mux_brutal_enable]:
                base['mux_brutal_enable'] = True
                base['mux_brutal_up_mbps'] = hconfigs.get(ConfigEnum.mux_brutal_up_mbps, 10)
                base['mux_brutal_down_mbps'] = hconfigs.get(ConfigEnum.mux_brutal_down_mbps, 10)

    if is_cdn and proxy.proto in {'vless', 'trojan', "vmess"}:
        if hconfigs[ConfigEnum.tls_fragment_enable]:
            base["tls_fragment_enable"] = True
            base["tls_fragment_size"] = hconfigs[ConfigEnum.tls_fragment_size]
            base["tls_fragment_sleep"] = hconfigs[ConfigEnum.tls_fragment_sleep]

        if hconfigs[ConfigEnum.tls_mixed_case]:
            base["tls_mixed_case"] = hconfigs[ConfigEnum.tls_mixed_case]
            base['host'] = hutils.random.random_case(base['host'])
            base['sni'] = hutils.random.random_case(base['sni'])
            base['server'] = hutils.random.random_case(base['server'])
            if base.get('fakedomain'):
                base['fakedomain'] = hutils.random.random_case(base['fakedomain'])

        if hconfigs[ConfigEnum.tls_padding_enable]:
            base["tls_padding_enable"] = hconfigs[ConfigEnum.tls_padding_enable]
            base["tls_padding_length"] = hconfigs[ConfigEnum.tls_padding_length]

    if "tcp" in proxy.transport:
        base['transport'] = 'tcp'
        base['path'] = f'/{path[base["proto"]]}{hconfigs[ConfigEnum.path_tcp]}'
        return base
    if proxy.transport in ["ws", "WS"]:
        base['transport'] = 'ws'
        base['path'] = f'/{path[base["proto"]]}{hconfigs[ConfigEnum.path_ws]}'
        base["host"] = domain
        return base

    if proxy.transport in [ProxyTransport.httpupgrade]:
        base['transport'] = 'httpupgrade'
        base['path'] = f'/{path[base["proto"]]}{hconfigs[ConfigEnum.path_httpupgrade]}'
        base["host"] = domain
        return base

    if proxy.transport == "grpc":
        base['transport'] = 'grpc'
        # base['grpc_mode'] = "multi" if hconfigs[ConfigEnum.core_type]=='xray' else 'gun'
        base['grpc_mode'] = 'gun'
        base['grpc_service_name'] = f'{path[base["proto"]]}{hconfigs[ConfigEnum.path_grpc]}'
        base['path'] = base['grpc_service_name']
        return base

    if "h1" in proxy.transport:
        base['transport'] = 'tcp'
        base['alpn'] = 'http/1.1'
        return base
    if ProxyProto.ssh == proxy.proto:
        base['private_key'] = g.account.ed25519_private_key
        base['host_key'] = hutils.proxy.get_ssh_hostkeys(False)
        # base['ssh_port'] = hconfig(ConfigEnum.ssh_server_port)
        return base
    return {'name': name, 'msg': 'not valid', 'type': 'error', 'proto': proxy.proto}


class ProxyJsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, IPv4Address) or isinstance(obj, IPv6Address):
            return str(obj)
        return super().default(obj)



from flask_admin.contrib.sqla import ModelView
from hiddifypanel import auth
from flask_admin.form import SecureForm


class AdminLTEModelView(ModelView):
    form_base_class = SecureForm
    edit_modal = True
    create_modal = True

    list_template = 'hiddify-flask-admin/list.html'
    create_template = 'flask-admin/model/create.html'
    edit_template = 'flask-admin/model/edit.html'
    details_template = 'flask-admin/model/details.html'

    create_modal_template = 'flask-admin/model/modals/create.html'
    edit_modal_template = 'flask-admin/model/modals/edit.html'
    details_modal_template = 'flask-admin/model/modals/details.html'

    # form_base_class = SecureForm
    def inaccessible_callback(self, name, **kwargs):
        return auth.redirect_to_login()  # type: ignore



from apiflask import abort
from flask.views import MethodView
from flask_babel import lazy_gettext as _
from flask import g
from loguru import logger

from hiddifypanel.models.child import Child
from hiddifypanel.auth import login_required
from hiddifypanel import hutils


class SyncWithParentApi(MethodView):
    decorators = [login_required(node_auth=True)]

    def post(self):
        logger.info(f"Syncing panel with parent called by {Child.node.unique_id}")
        if not hutils.node.child.sync_with_parent():
            logger.error("Sync with parent failed")
            abort(400, _('child.sync-failed'))  # type: ignore
        logger.success(f"Synced panel with parent {Child.node.unique_id}")
        return {'status': 200, 'msg': 'ok'}



# from hiddifypanel.models.domain import Domain
# from sqlalchemy.orm import backref
#

# from hiddifypanel.database import db

# ShowDomainParent = db.Table('show_domain_parent',
#                             db.Column('domain_id', db.Integer, db.ForeignKey('parent_domain.id'), primary_key=True),
#                             db.Column('related_id', db.Integer, db.ForeignKey('domain.id'), primary_key=True)
#                             )


# class ParentDomain(db.Model):
#     id = db.Column(db.Integer, primary_key=True, autoincrement=True)
#     domain = db.Column(db.String(200), nullable=False, unique=True)
#     alias = db.Column(db.String(200), nullable=False, unique=False)
#     show_domains = db.relationship('Domain', secondary=ShowDomainParent,
#                                    # primaryjoin=id==ShowDomainParent.c.domain_id,
#                                    # secondaryjoin=id==ShowDomainParent.c.related_id,
#                                    backref=backref('parent_domains', lazy='dynamic')
#                                    )

#     def to_dict(self):
#         return {
#             'domain': self.domain,
#             'alias': self.alias,
#             'show_domains': [dd.domain for dd in self.show_domains]
#         }

#     @staticmethod
#     def add_or_update(commit=True, **parent_domain):
#         dbdomain = ParentDomain.query.filter(ParentDomain.domain == parent_domain['domain']).first()
#         if not dbdomain:
#             dbdomain = ParentDomain(domain=parent_domain['domain'])
#             db.session.add(dbdomain)
#         show_domains = parent_domain.get('show_domains', [])
#         dbdomain.show_domains = Domain.query.filter(Domain.domain.in_(show_domains)).all()
#         dbdomain.alias = parent_domain.get('alias')
#         if commit:
#             db.session.commit()

#     @staticmethod
#     def bulk_register(parent_domains, commit=True, remove=False):
#         for p in parent_domains:
#             ParentDomain.add_or_update(commit=False, **p)
#         if remove:
#             dd = {p.domain: 1 for p in parent_domains}
#             for d in ParentDomain.query.all():
#                 if d.domain not in dd:
#                     db.session.delete(d)
#         if commit:
#             db.session.commit()



from apiflask import APIBlueprint

bp = APIBlueprint("api_child", __name__, url_prefix="/<proxy_path>/api/v2/child/", enable_openapi=True)


def init_app(app):
    with app.app_context():
        from .register_parent_api import RegisterWithParentApi
        from .sync_parent_api import SyncWithParentApi
        from .actions import ApplyConfig, Restart, Status, UpdateUsage, Install
        bp.add_url_rule('/sync-parent/', view_func=SyncWithParentApi)
        bp.add_url_rule('/register-parent/', view_func=RegisterWithParentApi)
        bp.add_url_rule('/status/', view_func=Status)
        bp.add_url_rule('/restart/', view_func=Restart)
        bp.add_url_rule('/apply-config/', view_func=ApplyConfig)
        bp.add_url_rule('/install/', view_func=Install)
        bp.add_url_rule('/update-usage/', view_func=UpdateUsage)
    app.register_blueprint(bp)



from . import network
from . import system
from . import importer
from . import github_issue
from . import flask
from . import convert
from . import random
from . import encode
from . import auth
from . import utils
from . import model
from . import crypto
from . import proxy
from . import node



from babel.dates import format_timedelta as babel_format_timedelta
from datetime import datetime
from datetime import timedelta
from flask import g

from hiddifypanel.models.config import hconfig
from hiddifypanel.models.config_enum import ConfigEnum
from flask_babel import gettext as _


def is_int(input: str) -> bool:
    try:
        int(input)
        return True
    except BaseException:
        return False


def to_int(s: str | None) -> int | None:
    '''Returns 0 if <s> is not a number'''
    if not s:
        return None
    try:
        return int(s)
    except BaseException:
        return 0


def date_to_json(d: datetime) -> str | None:
    return d.strftime("%Y-%m-%d") if d else None


def json_to_date(date_str: str) -> datetime | str:
    try:
        return datetime.strptime(date_str, '%Y-%m-%d')
    except BaseException:
        return date_str


def time_to_json(d: datetime) -> str | None:
    return __fix_time_format(d.strftime("%Y-%m-%d %H:%M:%S")) if d else None


def __fix_time_format(time_str):
    'Convert "1-00-00 00:00:00" to "0001-00-00 00:00:00"'
    t = time_str
    char_index = t.find('-')
    year_part = t[:char_index]

    if len(year_part) < 4:
        t = year_part.zfill(4) + t[char_index:]

    return t


def json_to_time(time_str: str) -> datetime | str:
    try:
        return datetime.strptime(__fix_time_format(time_str), "%Y-%m-%d %H:%M:%S")
    except BaseException:
        return time_str


def format_timedelta(delta: timedelta, add_direction: bool = True, granularity: str = "days") -> str:
    res = delta.days
    locale = g.locale if g and hasattr(g, "locale") else hconfig(ConfigEnum.admin_lang)
    if granularity == "days" and delta.days == 0:
        res = _("0 - Last day")
    elif delta.days < 7 or delta.days >= 60:
        res = babel_format_timedelta(delta, threshold=1, add_direction=add_direction, locale=locale)
    elif delta.days < 60:
        res = babel_format_timedelta(delta, granularity="day", threshold=10, add_direction=add_direction, locale=locale)
    return str(res)



from flask import request, g
# from hiddifypanel.cache import cache
from hiddifypanel.models import *

import flask_bootstrap
import hiddifypanel
from flask_babel import Babel
from flask_session import Session

import datetime

from dotenv import dotenv_values
import os
import sys
from apiflask import APIFlask
from werkzeug.middleware.proxy_fix import ProxyFix
from loguru import logger
from hiddifypanel.panel.init_db import init_db


def create_app(*args, cli=False, **config):

    app = APIFlask(__name__, static_url_path="/<proxy_path>/static/", instance_relative_config=True, version='2.0.0', title="Hiddify API",
                   openapi_blueprint_url_prefix="/<proxy_path>/api", docs_ui='elements', json_errors=False, enable_openapi=not cli)
    # app = Flask(__name__, static_url_path="/<proxy_path>/static/", instance_relative_config=True)

    if not cli:
        from hiddifypanel.cache import redis_client
        from hiddifypanel import auth
        app.config["PREFERRED_URL_SCHEME"] = "https"
        app.wsgi_app = ProxyFix(
            app.wsgi_app, x_for=1, x_proto=1, x_host=1, x_prefix=1,
        )
        app.servers = {
            'name': 'current',
            'url': '',
        }  # type: ignore
        app.info = {
            'description': 'Hiddify is a free and open source software. It is as it is.',
            'termsOfService': 'https://hiddify.com',
            'contact': {
                'name': 'API Support',
                'url': 'https://www.hiddify.com/support',
                'email': 'panel@hiddify.com'
            },
            'license': {
                'name': 'Creative Commons Zero v1.0 Universal',
                'url': 'https://github.com/hiddify/Hiddify-Manager/blob/main/LICENSE'
            }
        }
        # setup flask server-side session
        # app.config['APPLICATION_ROOT'] = './'
        # app.config['SESSION_COOKIE_DOMAIN'] = '/'
        app.config['SESSION_TYPE'] = 'redis'
        app.config['SESSION_REDIS'] = redis_client
        app.config['SESSION_PERMANENT'] = True
        app.config['PERMANENT_SESSION_LIFETIME'] = datetime.timedelta(days=10)
        Session(app)

        app.jinja_env.line_statement_prefix = '%'
        from hiddifypanel import hutils
        app.jinja_env.filters['b64encode'] = hutils.encode.do_base_64
        app.view_functions['admin.static'] = {}  # fix bug in apiflask
        flask_bootstrap.Bootstrap4(app)

    for c, v in dotenv_values(os.environ.get("HIDDIFY_CFG_PATH", 'app.cfg')).items():
        if v.isdecimal():
            v = int(v)
        else:
            v = True if v.lower() == "true" else (False if v.lower() == "false" else v)

        app.config[c] = v

    hiddifypanel.database.init_app(app)
    with app.app_context():
        init_db()

        # configure logger
        from hiddifypanel.models import ConfigEnum, hconfig
        logger.remove()
        log_format = '<green>{time:YYYY-MM-DD HH:mm:ss}</green> | <level>{level: <8}</level> | <cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - <level>{message}</level> | <level>{extra}</level>'
        logger.add(sys.stderr, format=log_format, level=hconfig(ConfigEnum.log_level), colorize=True, catch=True, enqueue=True, diagnose=False, backtrace=True)

    # flaskbabel = FlaskBabel(app)
    # @babel.localeselector

    def get_locale():
        # Put your logic here. Application can store locale in
        # user profile, cookie, session, etc.
        if "admin" in request.base_url:
            g.locale = auth.current_account.lang or hconfig(ConfigEnum.admin_lang) or 'fa'
        else:
            g.locale = auth.current_account.lang or hconfig(ConfigEnum.lang) or 'fa'
        return g.locale
    app.jinja_env.globals['get_locale'] = get_locale
    babel = Babel(app, locale_selector=get_locale)
    if not cli:
        hiddifypanel.panel.common.init_app(app)
        hiddifypanel.panel.common_bp.init_app(app)

        from hiddifypanel.panel import user, commercial, admin
        admin.init_app(app)
        user.init_app(app)
        commercial.init_app(app)

    app.config.update(config)  # Override with passed config
    # app.config['WTF_CSRF_CHECK_DEFAULT'] = False
    app.config['WTF_CSRF_ENABLED'] = False
    # app.config['BABEL_TRANSLATION_DIRECTORIES'] = '/workspace/Hiddify-Server/hiddify-panel/src/translations.i18n'

    # from flask_wtf.csrf import CSRFProtect

    # csrf = CSRFProtect(app)

    # @app.before_request
    # def check_csrf():
    # if "/admin/user/" in request.base_url:
    #     return
    # if "/admin/domain/" in request.base_url:
    #     return
    # if "/admin/actions/" in request.base_url:
    #     return
    # if "/api/" in request.base_url:
    #     return
    # csrf.protect()

    hiddifypanel.panel.cli.init_app(app)
    return app


def create_app_wsgi(*args, **kwargs):
    # workaround for Flask issue
    # that doesn't allow **config
    # to be passed to create_app
    # https://github.com/pallets/flask/issues/4170
    cli = ("hiddifypanel" in sys.argv[0]) or (sys.argv[1] in ["update-usage", "all-configs", "admin_links", "admin_path"])
    app = create_app(cli=cli)
    return app


# def create_cli_app(*args, **kwargs):
#     #     # workaround for Flask issue
#     #     # that doesn't allow **config
#     #     # to be passed to create_app
#     #     # https://github.com/pallets/flask/issues/4170
#     # print(kwargs)
#     app = create_app(*args, cli=True, **kwargs)
#     return app



# from hiddifypanel import hutils
# from hiddifypanel.auth import login_required
# from wtforms.validators import Regexp
# from hiddifypanel.models import *
# from wtforms.validators import Regexp, ValidationError
# from hiddifypanel.panel.admin.adminlte import AdminLTEModelView
# from flask_babel import gettext as __
# from flask_babel import lazy_gettext as _
# from hiddifypanel.panel import hiddify
# from flask import Markup, g  # type: ignore
# from . import ParentDomain

# # Define a custom field type for the related domains
# from flask_admin.form.fields import Select2TagsField, Select2Field
# from flask import current_app


# class ParentDomainAdmin(AdminLTEModelView):
#     column_hide_backrefs = False

#     list_template = 'model/domain_list.html'
#     # edit_modal = True
#     # form_overrides = {'work_with': Select2Field}

#     column_descriptions = dict(
#         domain=_("domain.description"),
#         alias=_('The name shown in the configs for this domain.'),
#         show_domains=_('You can select the configs with which domains show be shown in the user area. If you select all, automatically, all the new domains will be added for each users.')
#         # current_usage_GB="in GB"
#     )

#     # create_modal = True
#     can_export = False
#     form_widget_args = {'show_domains': {'class': 'form-control ltr'}}
#     form_args = {

#         'domain': {
#             'validators': [Regexp(r'^([A-Za-z0-9\-\.]+\.[a-zA-Z]{2,})$', message=__("Should be a valid domain"))]
#         },

#     }
#     column_list = ["domain", "alias", "show_domains_list"]
#     # column_editable_list=["domain"]
#     # column_filters=["domain","mode"]
#     # form_excluded_columns=['work_with']
#     column_searchable_list = ["domain"]
#     column_labels = {
#         "domain": _("domain.domain"),
#         # "mode": _("domain.mode"),
#         'alias': _('Alias'),
#         'show_domains': _('Show Domains'),
#         'show_domains_list': _('Show Domains')
#     }

#     form_columns = ['domain', "alias", 'show_domains']

#     def _domain_admin_link(view, context, model, name):
#         admin_link = hiddify.get_account_panel_link(g.account, model.domain)
#         return Markup(f'<div class="btn-group"><a href="{admin_link}" class="btn btn-xs btn-secondary">' + _("admin link") +
#                       f'</a><a href="{admin_link}" class="btn btn-xs btn-info ltr" target="_blank">{model.domain}</a></div>')

#     def _domain_ip(view, context, model, name):
#         dip = hutils.network.get_domain_ip(model.domain)
#         myip = hutils.network.get_ip_str(4)
#         if myip == dip and model.mode == DomainType.direct:
#             badge_type = ''
#         elif dip and model.mode != DomainType.direct and myip != dip:
#             badge_type = 'warning'
#         else:
#             badge_type = 'danger'
#         return Markup(f'<span class="badge badge-{badge_type}">{dip}</span>')

#     def _show_domains(view, context, model, name):
#         # return Markup(f'<span class="badge badge-{badge_type}">{dip}</span>')
#         res = ""
#         for d in model.show_domains:
#             res += f'<span class="badge">{d.domain}</span>'
#         return Markup(res)

#     column_formatters = {
#         # 'domain_ip': _domain_ip,
#         'domain': _domain_admin_link,
#         'show_domains_list': _show_domains
#     }

#     def search_placeholder(self):
#         return f"{_('search')} {_('domain.domain')}"

#     # def on_form_prefill(self, form, id):
#         # Get the Domain object being edited
#         # domain = self.session.query(Domain).get(id)

#         # Pre-select the related domains in the checkbox list
#         # form.show_domains = [d.id for d in Domain.query.all()]

#     def on_model_change(self, form, model, is_created):
#         model.domain = model.domain.lower()

#         dip = hutils.network.get_domain_ip(model.domain)
#         if dip == None:
#             raise ValidationError(
#                 _("Domain can not be resolved! there is a problem in your domain"))
#         # if not hutils.network.check_connection_for_domain(model.domain):
#         #     raise ValidationError(
#         #         _("Domain is not correctly mapped to this server!"))

#         if len(model.show_domains) == Domain.query.count():
#             model.show_domains = []

#         hutils.flask.flash_config_success(restart_mode=ApplyMode.apply, domain_changed=True)

#     def on_model_delete(self, model):
#         if len(ParentDomain.query.all()) <= 1:
#             raise ValidationError(f"at least one domain should exist")
#         hutils.flask.flash_config_success(restart_mode=ApplyMode.apply, domain_changed=True)

#     def is_accessible(self):
#         if login_required(roles={Role.super_admin, Role.admin})(lambda: True)() != True:
#             return False
#         return True



#!/usr/bin/env python3
import os
import urllib.request

from flask_classful import FlaskView, route
from flask import render_template, request, redirect, g
from hiddifypanel.hutils.flask import hurl_for
from hiddifypanel.auth import login_required
from flask import current_app as app
from flask_babel import gettext as _


from hiddifypanel import hutils
from hiddifypanel.models import *
from hiddifypanel.panel import hiddify, usage
from hiddifypanel.panel.run_commander import commander, Command


class Actions(FlaskView):

    @login_required(roles={Role.super_admin})
    def index(self):
        return render_template('index.html')

    @login_required(roles={Role.super_admin})
    def viewlogs(self):
        log_files = hutils.flask.list_dir_files(f"{app.config['HIDDIFY_CONFIG_PATH']}log/system/")
        return render_template('view_logs.html', log_files=log_files)

    @login_required(roles={Role.super_admin})
    @route('apply_configs', methods=['POST'])
    def apply_configs(self):
        return self.reinstall(False)

    @route('reset', methods=['POST'])
    @login_required(roles={Role.super_admin})
    def reset(self):
        return self.reset2()

    @login_required(roles={Role.super_admin})
    def reset2(self):
        res = render_template("result.html",
                              out_type="info",
                              out_msg="",
                              log_file_url=get_log_api_url(),
                              log_file='restart.log',
                              show_success=True,
                              domains=get_domains())

        # run restart.sh
        commander(Command.restart_services)

        return res

    @login_required(roles={Role.super_admin})
    @route('reinstall', methods=['POST'])
    def reinstall(self, complete_install=True, domain_changed=False):
        return self.reinstall2(complete_install, domain_changed)

    @login_required(roles={Role.super_admin})
    def reinstall2(self, complete_install=True, domain_changed=False):
        if int(hconfig(ConfigEnum.db_version)) < 9:
            return ("Please update your panel before this action.")
        if hutils.node.is_child():
            if not hutils.node.child.sync_with_parent():
                hutils.flask.flash(_('child.sync-failed'), 'danger')  # type: ignore

        domain_changed = request.args.get("domain_changed", str(domain_changed)).lower() == "true"
        complete_install = request.args.get("complete_install", str(complete_install)).lower() == "true"
        if domain_changed:
            hutils.flask.flash((_('Your domains changed. Please do not forget to copy admin links, otherwise you can not access to the panel anymore.')), 'info')
        # hutils.flask.flash(f'complete_install={complete_install} domain_changed={domain_changed} ', 'info')
        # return render_template("result.html")
        # hiddify.add_temporary_access()
        file = "install.sh" if complete_install else "apply_configs.sh"
        try:
            server_ip = urllib.request.urlopen('https://v4.ident.me/').read().decode('utf8')
        except BaseException:
            server_ip = "server_ip"

        admin_links = f"<h5 >{_('Admin Links')}</h5><ul>"

        admin_links += f"<li><span class='badge badge-danger'>{_('Not Secure')}</span>: <a class='badge ltr share-link' href='{hiddify.get_account_panel_link(g.account, server_ip,is_https=False)}'>{hiddify.get_account_panel_link(g.account, server_ip,is_https=False)}</a></li>"
        domains = get_panel_domains()
        # domains=[*domains,f'{server_ip}.sslip.io']

        for d in domains:
            link = hiddify.get_account_panel_link(g.account, d)
            admin_links += f"<li><a target='_blank' class='badge ltr' href='{link}'>{link}</a></li>"

        resp = render_template("result.html",
                               out_type="info",
                               out_msg=_("Success! Please wait around 4 minutes to make sure everything is updated. During this time, please save your proxy links which are:") +
                               admin_links,
                               log_file_url=get_log_api_url(),
                               log_file="0-install.log",
                               show_success=True,
                               domains=get_domains())

        # subprocess.Popen(f"sudo {config['HIDDIFY_CONFIG_PATH']}/{file} --no-gui".split(" "), cwd=f"{config['HIDDIFY_CONFIG_PATH']}", start_new_session=True)

        # run install.sh or apply_configs.sh
        commander(Command.install if complete_install else Command.apply)

        # import time
        # time.sleep(1)
        return resp

    @login_required(roles={Role.super_admin})
    def change_reality_keys(self):
        key = hutils.crypto.generate_x25519_keys()
        set_hconfig(ConfigEnum.reality_private_key, key['private_key'])
        set_hconfig(ConfigEnum.reality_public_key, key['public_key'])
        hutils.flask.flash_config_success(restart_mode=ApplyMode.restart, domain_changed=False)
        return redirect(hurl_for('admin.SettingAdmin:index'))

    @ login_required(roles={Role.super_admin})
    def status(self):
        # run status.sh
        commander(Command.status)
        return render_template("result.html",
                               out_type="info",
                               out_msg=_("see the log in the bellow screen"),
                               log_file_url=get_log_api_url(),
                               log_file="status.log",
                               show_success=False,
                               domains=get_domains())

    @ route('update', methods=['POST'])
    @ login_required(roles={Role.super_admin})
    def update(self):
        return self.update2()

    def update2(self):
        # hiddify.add_temporary_access()
        # run update.sh

        commander(Command.update)

        return render_template("result.html",
                               out_type="success",
                               out_msg=_("Success! Please wait around 5 minutes to make sure everything is updated."),
                               show_success=True,
                               log_file_url=get_log_api_url(),
                               log_file="update.log",
                               domains=get_domains())

    def get_some_random_reality_friendly_domain(self):
        test_domain = request.args.get("test_domain")
        import ping3
        from hiddifypanel.hutils.network.auto_ip_selector import IPASN, IPCOUNTRY
        ipv4 = hutils.network.get_ip_str(4)
        server_country = (IPCOUNTRY.get(ipv4) or {}).get('country', {}).get('iso_code', 'unknown')
        server_asn = (IPASN.get(ipv4) or {}).get('autonomous_system_organization', 'unknown')
        res = "<table><tr><th>Domain</th><th>IP</th><th>Country</th><th>ASN</th><th>Ping (ms)</th><th>TCP ping (ms)</th></tr>"
        res += f"<tr><td>Your Server</td><td>{ipv4}</td><td>{server_country}</td><td>{server_asn}</td><td>0</td></tr>"
        import time
        start = time.time()
        for d in [test_domain, *hutils.network.get_random_domains(30)]:
            if not d:
                continue
            if time.time() - start > 20:
                break

            tcp_ping = hutils.network.is_domain_reality_friendly(d)
            if tcp_ping:
                dip = str(hutils.network.get_domain_ip(d))
                dip_country = (IPCOUNTRY.get(dip) or {}).get('country', {}).get('iso_code', 'unknown')
                if dip_country == "IR":
                    continue
                response_time = -1
                try:
                    response_time = ping3.ping(d, unit='ms')
                    if response_time:
                        response_time = int(response_time)
                except BaseException:
                    pass
                dip_asn = (IPASN.get(dip) or {}).get('autonomous_system_organization', 'unknown')
                res += f"<tr><td>{d}</td><td>{dip}</td><td>{dip_country}</td><td>{dip_asn}</td><td>{response_time}</td><td>{tcp_ping}<td></tr>"

        return res + "</table>"

    @ login_required(roles={Role.super_admin})
    def update_usage(self):

        import json

        return render_template("result.html",
                               out_type="info",
                               out_msg=f'<pre class="ltr">{json.dumps(usage.update_local_usage(),indent=2)}</pre>',
                               log_file_url=None
                               )


def get_log_api_url():
    return f'/{g.get("new_proxy_path",g.proxy_path)}/api/v2/admin/log/'


def get_domains():
    return [str(d.domain).replace("*", hutils.random.get_random_string(3, 6)) for d in get_panel_domains(always_add_all_domains=True, always_add_ip=False)]



from typing import Optional
from hiddifypanel.models.config_enum import ConfigEnum, LogLevel, PanelMode, Lang
from flask import g

from hiddifypanel import Events
from hiddifypanel.database import db
from hiddifypanel.cache import cache
from hiddifypanel.models.child import Child, ChildMode
from sqlalchemy import Column, String, Boolean, Enum, ForeignKey, Integer
from strenum import StrEnum
from sqlalchemy_serializer import SerializerMixin


def error(st):
    from hiddifypanel.hutils.utils import error as err
    err(st)


class BoolConfig(db.Model, SerializerMixin):
    child_id = Column(Integer, ForeignKey('child.id'), primary_key=True, default=0)
    # category = db.Column(db.String(128), primary_key=True)
    key = Column(Enum(ConfigEnum), primary_key=True)
    value = Column(Boolean)

    def to_dict(d):
        return {
            'key': str(d.key),
            'value': d.value,
            'child_unique_id': d.child.unique_id if d.child else ''
        }

    @staticmethod
    def from_schema(schema):
        return schema.dump(BoolConfig())

    def to_schema(self):
        conf_dict = self.to_dict()
        from hiddifypanel.panel.commercial.restapi.v2.parent.schema import HConfigSchema
        return HConfigSchema().load(conf_dict)


class StrConfig(db.Model, SerializerMixin):
    child_id = Column(Integer, ForeignKey('child.id'), primary_key=True, default=0)
    # category = db.Column(db.String(128), primary_key=True)
    key = Column(Enum(ConfigEnum), primary_key=True, default=ConfigEnum.admin_secret)
    value = Column(String(2048))

    def to_dict(d):
        return {
            'key': str(d.key),
            'value': d.value,
            'child_unique_id': d.child.unique_id if d.child else ''
        }

    @staticmethod
    def from_schema(schema):
        return schema.dump(StrConfig())

    def to_schema(self):
        conf_dict = self.to_dict()
        from hiddifypanel.panel.commercial.restapi.v2.parent.schema import HConfigSchema
        return HConfigSchema().load(conf_dict)


@cache.cache(ttl=500)
def hconfig(key: ConfigEnum, child_id: Optional[int] = None):  # -> str | int | StrEnum | None:
    if child_id is None:
        child_id = Child.current().id

    value = None
    try:
        if key.type == bool:
            bool_conf = BoolConfig.query.filter(BoolConfig.key == key, BoolConfig.child_id == child_id).first()
            if bool_conf:
                value = bool_conf.value
            else:
                error(f'bool {key} not found ')
        else:
            str_conf = StrConfig.query.filter(StrConfig.key == key, StrConfig.child_id == child_id).first()
            if str_conf:
                value = str_conf.value
            else:
                error(f'str {key} not found ')
    except BaseException:
        error(f'{key} error!')
        raise
    if value != None:
        if key.type == int:
            return int(value)
        elif hasattr(key.type, 'from_str'):
            return key.type.from_str(value)

    return value


def set_hconfig(key: ConfigEnum, value: str | int | bool, child_id: int | None = None, commit: bool = True):
    if child_id is None:
        child_id = Child.current().id

    print(f"chainging .... {key}---{value}---{child_id}---{commit}")
    if key.type == int and value != None:
        int(value)  # for testing int

    # hconfig.invalidate(key, child_id)
    # get_hconfigs.invalidate(child_id)
    hconfig.invalidate(key, child_id)
    hconfig.invalidate(key, child_id=child_id)
    hconfig.invalidate(key=key, child_id=child_id)
    if child_id == Child.current().id:
        hconfig.invalidate(key)
    # hconfig.invalidate_all()
    get_hconfigs.invalidate_all()
    old_v = None
    if key.type == bool:
        dbconf = BoolConfig.query.filter(BoolConfig.key == key, BoolConfig.child_id == child_id).first()
        if not dbconf:
            dbconf = BoolConfig(key=key, value=value, child_id=child_id)
            db.session.add(dbconf)
        else:
            old_v = dbconf.value
    else:
        dbconf = StrConfig.query.filter(StrConfig.key == key, StrConfig.child_id == child_id).first()
        if not dbconf:
            dbconf = StrConfig(key=key, value=value, child_id=child_id)
            db.session.add(dbconf)
        else:
            old_v = dbconf.value
    dbconf.value = value
    error(f"changing {key} from {old_v} to {value}")
    Events.config_changed.notify(conf=dbconf, old_value=old_v)

    if child_id == 0 and key.hide_in_virtual_child:
        for child in Child.query.filter(Child.mode == ChildMode.virtual, Child.id != 0).all():
            set_hconfig(key, value, child.id)

    if commit:
        db.session.commit()


@cache.cache(ttl=500,)
def get_hconfigs(child_id: int | None = None, json=False) -> dict:
    if child_id is None:
        child_id = Child.current().id

    return {**{f'{u.key}' if json else u.key: u.value for u in BoolConfig.query.filter(BoolConfig.child_id == child_id).all() if u.key.type == bool},
            **{f'{u.key}' if json else u.key: int(u.value) if u.key.type == int and u.value != None else u.value for u in StrConfig.query.filter(StrConfig.child_id == child_id).all() if u.key.type != bool},
            # ConfigEnum.telegram_fakedomain:hdomain(DomainType.telegram_faketls),
            # ConfigEnum.ssfaketls_fakedomain:hdomain(DomainType.ss_faketls),
            # ConfigEnum.fake_cdn_domain:hdomain(DomainType.fake_cdn)
            }


def get_hconfigs_childs(child_ids: list[int], json=False):
    if len(child_ids) == 0:
        child_ids = [c.id for c in Child.query.all()]
    return {c: get_hconfigs(c, json) for c in child_ids}


def add_or_update_config(commit: bool = True, child_id: int | None = None, override_unique_id: bool = True, **config):
    if child_id is None:
        child_id = Child.current().id
    c = config['key']
    ckey = ConfigEnum(c)
    if c == ConfigEnum.unique_id and not override_unique_id:
        return

    v = str(config['value']).lower() == "true" if ckey.type == bool else config['value']
    if ckey in [ConfigEnum.db_version]:
        return
    set_hconfig(ckey, v, child_id, commit=commit)


def bulk_register_configs(hconfigs, commit: bool = True, froce_child_unique_id: str | None = None, override_unique_id: bool = True):
    from hiddifypanel.panel import hiddify
    for conf in hconfigs:
        # print(conf)
        if conf['key'] == ConfigEnum.unique_id and not override_unique_id:
            continue
        child_id = hiddify.get_child(unique_id=froce_child_unique_id)
        # print(conf, child_id, conf.get('child_unique_id', None), override_child_unique_id)
        add_or_update_config(commit=False, child_id=child_id, **conf)
    if commit:
        db.session.commit()



from typing import List, Literal, Tuple, Union
from urllib.parse import urlparse
import urllib.request
import ipaddress
from hiddifypanel.hutils.network.auto_ip_selector import IPASN
import netifaces
import requests
import random
import socket
import time
import ssl
import re
import os


from hiddifypanel.models import *
from hiddifypanel.cache import cache


def get_domain_ip(domain: str, retry: int = 3, version: Literal[4, 6] | None = None) -> Union[ipaddress.IPv4Address, ipaddress.IPv6Address, None]:
    res = None
    if not version:
        try:
            res = socket.gethostbyname(domain)
        except BaseException:
            pass

    if not res and version != 6:
        try:
            res = socket.getaddrinfo(domain, None, socket.AF_INET)[0][4][0]
        except BaseException:
            pass

    if not res and version != 4:
        try:
            res = f"[{socket.getaddrinfo(domain, None, socket.AF_INET6)[0][4][0]}]"

            res = res[1:-1]

        except BaseException:
            pass

    if retry <= 0 or not res:
        return None

    return ipaddress.ip_address(res) or get_domain_ip(domain, retry=retry - 1) if res else None


def get_socket_public_ip(version: Literal[4, 6]) -> Union[ipaddress.IPv4Address, ipaddress.IPv6Address, None]:
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        if version == 6:
            s.connect(("2001:4860:4860::8888", 80))
        else:
            s.connect(("8.8.8.8", 80))
        ip_address = ipaddress.ip_address(s.getsockname()[0])
        s.close()
        return ip_address if ip_address.is_global else None
    except socket.error:
        return None


def get_interface_public_ip(version: Literal[4, 6]) -> List[Union[ipaddress.IPv4Address, ipaddress.IPv6Address]]:
    addresses = []
    try:
        interfaces = netifaces.interfaces()
        for interface in interfaces:
            if version == 4:
                address_info = netifaces.ifaddresses(interface).get(netifaces.AF_INET, [])
            elif version == 6:
                address_info = netifaces.ifaddresses(interface).get(netifaces.AF_INET6, [])
            else:
                continue

            if address_info:
                for addr in address_info:
                    address = ipaddress.ip_address(addr['addr'])
                    if address.is_global:
                        addresses.append(address)

        return addresses

    except (OSError, KeyError):
        return []


@cache.cache(ttl=600)
def get_ips(version: Literal[4, 6]) -> List[Union[ipaddress.IPv4Address, ipaddress.IPv6Address]]:
    addrs = []
    i_ips = get_interface_public_ip(version)
    if i_ips:
        addrs = i_ips

    s_ip = get_socket_public_ip(version)
    if s_ip:
        addrs.append(s_ip)

    # send request
    try:
        ip = urllib.request.urlopen(f'https://v{version}.ident.me/').read().decode('utf8')
        if ip:
            addrs.append(ipaddress.ip_address(ip))
    except BaseException:
        pass

    # remove duplicates
    return list(set(addrs))


@cache.cache(ttl=600)
def get_ip_str(version: Literal[4, 6], retry: int = 5) -> str | None:
    ip = get_ip(version, retry)
    if ip is None:
        return None
    return str(ip)


@cache.cache(ttl=600)
def get_ip(version: Literal[4, 6], retry: int = 5) -> ipaddress.IPv4Address | ipaddress.IPv6Address | None:
    ips = get_interface_public_ip(version)
    ip = None
    if ips:
        ip = random.sample(ips, 1)[0]

    if ip is None:
        ip = get_socket_public_ip(version)

    if ip is None:
        try:
            ip = urllib.request.urlopen(f'https://v{version}.ident.me/').read().decode('utf8')
            if ip:
                ip = ipaddress.ip_address(ip)
        except BaseException:
            pass
    if ip is None and retry > 0:
        ip = get_ip(version, retry=retry - 1)
    return ip


def check_connection_to_remote(api_url: str) -> bool:

    path = f"{api_url}/api/v1/hello/"

    try:
        _ = requests.get(path, verify=False, timeout=2).json()
        return True

    except BaseException:
        return False


def check_connection_for_domain(domain: str) -> bool:

    proxy_path = hconfig(ConfigEnum.proxy_path_admin)
    admin_secret = hconfig(ConfigEnum.admin_secret)
    path = f"{proxy_path}/{admin_secret}/api/v1/hello/"
    try:
        print(f"https://{domain}/{path}")
        res = requests.get(
            f"https://{domain}/{path}", verify=False, timeout=10).json()
        return res['status'] == 200

    except BaseException:
        try:
            print(f"http://{domain}/{path}")
            res = requests.get(
                f"http://{domain}/{path}", verify=False, timeout=10).json()
            return res['status'] == 200
        except BaseException:
            try:
                print(f"http://{get_domain_ip(domain)}/{path}")
                res = requests.get(
                    f"http://{get_domain_ip(domain)}/{path}", verify=False, timeout=10).json()
                return res['status'] == 200
            except BaseException:
                return False


def get_random_domains(count: int = 1, retry: int = 3) -> List[str]:
    try:
        irurl = "https://api.ooni.io/api/v1/measurements?probe_cc=IR&test_name=web_connectivity&anomaly=false&confirmed=false&failure=false&order_by=test_start_time&limit=1000"
        # cnurl="https://api.ooni.io/api/v1/measurements?probe_cc=CN&test_name=web_connectivity&anomaly=false&confirmed=false&failure=false&order_by=test_start_time&limit=1000"
        data_ir = requests.get(irurl).json()
        # data_cn=requests.get(url).json()

        domains = [urlparse(d['input']).netloc.lower() for d in data_ir['results'] if d['scores']['blocking_country'] == 0.0]
        domains = [d for d in domains if not d.endswith(".ir") and ".gov" not in d]

        return random.sample(domains, count)
    except Exception as e:
        print('Error, getting random domains... ', e, 'retrying...', retry)
        if retry <= 0:
            defdomains = ["fa.wikipedia.org", 'en.wikipedia.org', 'wikipedia.org', 'yahoo.com', 'en.yahoo.com']
            print('Error, using default domains')
            return random.sample(defdomains, count)
        return get_random_domains(count, retry - 1)


# not used
def is_domain_support_tls_13(domain: str) -> bool:
    context = ssl.create_default_context()
    port = 433
    with socket.create_connection((domain, port)) as sock:
        with context.wrap_socket(sock, server_hostname=domain) as ssock:
            return ssock.version() == "TLSv1.3"


def is_domain_support_h2(sni: str, server: str = '') -> bool:
    try:

        context = ssl.create_default_context(purpose=ssl.Purpose.SERVER_AUTH)
        context.options |= (ssl.OP_NO_SSLv2 | ssl.OP_NO_SSLv3 | ssl.OP_NO_TLSv1 | ssl.OP_NO_TLSv1_1)
        context.options |= ssl.OP_NO_COMPRESSION
        context.set_ciphers("ECDHE+AESGCM:ECDHE+CHACHA20:DHE+AESGCM:DHE+CHACHA20")
        context.set_alpn_protocols(["h2"])
        start_time = time.monotonic()
        with socket.create_connection((server or sni, 443), timeout=2) as sock:

            with context.wrap_socket(sock, server_hostname=sni) as ssock:
                elapsed_time = time.monotonic() - start_time
                valid = ssock.version() == "TLSv1.3"
                if valid:
                    if int(max(1, elapsed_time * 1000)):
                        return True
                return False
    except Exception as e:
        print(f'{sni} {e}')
        return False


def is_domain_reality_friendly(domain: str) -> bool:
    return is_domain_support_h2(domain)


def fallback_domain_compatible_with_servernames(fallback_domain: str, servername: str) -> bool:
    return is_domain_support_h2(servername, fallback_domain)


def get_random_decoy_domain() -> str:
    for _ in range(10):
        domains = get_random_domains(10)
        for d in domains:
            if is_domain_use_letsencrypt(d):
                return d

    return "bbc.com"


def is_domain_use_letsencrypt(domain: str) -> bool:
    """
    This function is used to filter the payment and big companies to
    avoid phishing detection
    """
    try:
        # Create a socket connection to the website
        with socket.create_connection((domain, 443)) as sock:
            context = ssl.create_default_context()
            with context.wrap_socket(sock, server_hostname=domain) as ssock:
                certificate = ssock.getpeercert()

        issuer = dict(x[0] for x in certificate.get("issuer", []))

        return issuer['organizationName'] == "Let's Encrypt"
    except BaseException:
        return False


@cache.cache(ttl=300)
def get_direct_host_or_ip(prefer_version: int) -> str:
    from hiddifypanel.models import Domain
    direct = Domain.query.filter(Domain.mode == DomainType.direct, Domain.sub_link_only == False).first()
    if not direct:
        direct = Domain.query.filter(Domain.mode == DomainType.direct).first()
    if direct:
        return direct.domain

    direct = get_ip_str(prefer_version)
    if direct:
        return direct

    return get_ip_str(4 if prefer_version == 6 else 6)


# not used
def get_warp_info() -> str:
    proxies = dict(http='socks5://127.0.0.1:3000',
                   https='socks5://127.0.0.1:3000')
    res = requests.get("https://cloudflare.com/cdn-cgi/trace", proxies=proxies, timeout=1).text

    dicres = {line.split("=")[0]: line.split("=")[0] for line in res}
    return str(dicres)


def is_ssh_password_authentication_enabled() -> bool:
    if os.path.isfile('/etc/ssh/sshd_config'):
        with open('/etc/ssh/sshd_config', 'r') as f:
            for line in f.readlines():
                line = line.strip()
                if line.startswith('#'):
                    continue
                if re.search("^PasswordAuthentication\\s+no", line, re.IGNORECASE):
                    return False

    return True


def is_out_of_range_port(port: int) -> bool:
    return port < 1 or port > 65535


def add_number_to_ipv4(ip: str, number: int) -> str:
    octets = list(map(int, ip.split('.')))

    octets[2] = (octets[2] + (octets[3] + number) // 256)
    octets[3] = (octets[3] + number) % 256

    return f"{octets[0]}.{octets[1]}.{octets[2]}.{octets[3]}"


def add_number_to_ipv6(ip: str, number: int) -> str:
    segments = ip.split(':')

    # Increment the last segment by the specified number
    segments[-1] = hex(int(segments[-1] or "0", 16) + number)[2:]

    # Join the segments back together with colons
    modified_ipv6 = ":".join(segments)

    return modified_ipv6


def is_in_same_asn(domain_or_ip: str, domain_or_ip_target: str) -> bool:
    '''Returns True if domain is in panel ASN'''
    if not IPASN:
        return False
    try:
        ip = domain_or_ip if is_ip(domain_or_ip) else get_domain_ip(domain_or_ip)
        ip_target = domain_or_ip_target if is_ip(domain_or_ip_target) else get_domain_ip(domain_or_ip_target)

        if not ip or not ip_target:
            return False

        ip_asn = get_ip_asn_name(ip)
        ip_target_asn = get_ip_asn_name(ip_target)

        if not ip_asn or not ip_target_asn:
            return False

        return ip_asn == ip_target_asn
    except Exception as e:
        print(f"An error occurred: {e}")
        return False

        # hutils.flask.flash(_("selected domain for REALITY is not in the same ASN. To better use of the protocol, it is better to find a domain in the same ASN.") +
        #                    f"<br> Server ASN={asn_ipv4.get('autonomous_system_organization','unknown')}<br>{domain}_ASN={asn_dip.get('autonomous_system_organization','unknown')}", "warning")


def get_ip_asn_name(ip: ipaddress.IPv4Address | ipaddress.IPv6Address | str) -> str:
    try:
        if asn := IPASN.get(str(ip)):
            return str(asn.get('autonomous_system_organization', ''))
        return ''
    except:
        return ''


def is_ip(input: str):
    try:
        _ = ipaddress.ip_address(input)
        return True
    except:
        return False



from flask.views import MethodView
from flask import current_app as app, g
from apiflask import abort
from hiddifypanel.auth import login_required
from hiddifypanel.models.role import Role
from hiddifypanel.panel import hiddify
from hiddifypanel.drivers import user_driver
from hiddifypanel.models import User
from .user_api import UserSchema
from . import has_permission


class UsersApi(MethodView):
    decorators = [login_required({Role.super_admin, Role.admin, Role.agent})]

    @app.output(UserSchema(many=True))  # type: ignore
    def get(self):
        users = User.query.filter(User.added_by.in_(g.account.recursive_sub_admins_ids())).all() or abort(404, "You have no user")
        return [user.to_dict(False) for user in users]  # type: ignore

    @app.input(UserSchema, arg_name="data")  # type: ignore
    @app.output(UserSchema)  # type: ignore
    def put(self, data):
        uuid = data.get('uuid') or abort(422, "Parameter issue: 'uuid'")
        user = User.by_uuid(uuid)  # type: ignore

        if not data.get('added_by_uuid'):
            data['added_by_uuid'] = g.account.uuid

        # update check permission
        if user:
            if not has_permission(user):
                abort(403, "You don't have permission to access this user")
        User.add_or_update(**data)  # type: ignore

        dbuser = User.by_uuid(data['uuid']) or abort(502, "Unknown issue: User is not added")
        user_driver.add_client(dbuser)
        hiddify.quick_apply_users()
        return dbuser.to_dict(False)  # type: ignore
